# Image-based Many-language Programming Language Identification - Replication Package

**Version: 1.0**

This document details the steps needed to replicate the findings of the paper:
Francesca Del Bonifro, Maurizio Gabbrielli, Antonio Lategano, Stefano
Zacchiroli, *Image-based Many-language Programming Language Identification*,
PeerJ Computer Science, 2021.

The machine-learning models detailed in the paper are able to recognize the
programming languages using in short snippets of code rendered as images.

The results presented in the paper can be reproduced jumping directly to
section [Preprocessed artifacts](#preprocessed-artifacts) after installing the
needed software dependencies.

## Dependencies

The following software dependencies are required as prerequisites:

* [Python 3.6.8](https://www.python.org/)
* [Linguist](https://github.com/github/linguist)

Then you should install additional dependencies using `pip` as follows:

```
$ pip install -r requirements.txt
```

## Dataset

The starting
[dataset](https://annex.softwareheritage.org/public/dataset/content-samples/2019-10-08-github-top1k/)
is provided by [Software Heritage](https://www.softwareheritage.org) and not
included in this replication package due to size constraints. It contains
source code files extracted from the top 1000 most popular GitHub repositories
as determined by repository "stars" as of 2019-10-08. The initial dataset
contains two files:

* a `.tar` file that contains all source code files, identified by a SHA1
  cryptographic checksum. It needs to be decompressed in the `dataset/` folder.

* a `.gz` file that contains the information about the extension related to
  every single file present in the .tar. It needs to be placed in the
  `files_list/` folder.


### Preprocessing

Before model training and evaluation we need to generate images representing
source code snippets from the files in the initial dataset. These images will
then be used to train the classifiers. To do so, we need to preprocess the
files in the dataset.

#### Deduplicate, find frequencies

The `preprocess.py` file contains the logic for:

* finding **duplicates** in the dataset, i.e., files that share the same SHA1
  digest
* find the number of files and the **relative frequency** for every extension
* select **non-rare extensions**, i.e., extensions that have a frequency higher
  than 10^-5

To run the Python script provide the `.txt` file containing the initial list of
files (previously extracted in the `files_lists/` folder, as follows:

```
$ python src/preprocess.py files_list/filenames-utf8.txt
```

#### Ground truth

To establish a *ground truth* we use the
[Linguist](https://github.com/github/linguist) language recognition tool
developed by GitHub. Since GitHub doesn't provide a Python library for it, we
created a Python command-line wrapper to invoke Linguist.

To run the script use the following command. Note that it is parallelized to
use all available CPUs. Results will be stored in the `linguist_dataset/`
folder.

```
$ python src/linguist_preprocess.py
```

#### Generate source code bundles

To generate the data used to render code snippet to images we collect all the
files available for each language, and select from them a sample of at most
1000 files. Then we create for each language a single text file, called a
language "bundle", concatenating together selected snippets. To avoid
unbalances bundles that are too short are discarded. The bundle dataset is
produced by the following Python script invocation, which will place them in
the `bundles/` folder.

```
$ python src/build_dataset.py
```

## Image rendering

Code snippet images are rendered from bundles files by the Python script
invocation below. The scripts takes "windows" of 32 lines from bundles and
render them into a white-on-black.

The first part of the bundle file is used for the **train** and the
**validation** sets and allow text overlapping between windows. The last part
of bundles is used to render images for the **test set**.

```
$ python src/rendering.py
```

### Scramble

To understand to what extent the models are able to recognize language features
we have also created datasets that introduce bias into images by randomly
scrambling specific character classes.  Scrambling is applied only to the test
set.

Scrambled datasets are generated by the following script:

```
$ python src/rendering_scramble.py
```

## Preprocessed artifacts (optional)

While both bundle generation and image rendering can be rerun starting from the
initial dataset, we also include them in the replication package. That way
experiments can be rerun also without starting from the initial dataset, if so
desired.

Source code bundles can be found in the `bundles.tar.gz` archive. They can be
extracted as follows:

```
$ tar -xf bundles.tar.gz -C bundles
```

Rendered code snippets can be found in the `renderings.tar` archive, which has
been split into several parts (`renderings.tar.part*`) for convenience. To
reconstruct the renderings archive and extract it:

```
$ cat renderings.tar.part* > renderings.tar
$ tar -xf renderings.tar -C renderings
```

To *re*render new images from language bundles, extract the `bundles.tar.gz`
archive in the `bundles` folder and then proceed as discussed in the [Image
rendering](#image-rendering) section.


## Model

We tested three different generic image-classification models (ResNet34,
MobileNetv2 and AlexNet), pre-trained on ImageNet, and we applied transfer
learning to adapt them to the task of image-based programming language
identification.

After rendering the images, the performance measures of the models can be
obtained by running the following [Jupyter](https://jupyter.org/) notebooks:

* [resnet34.ipynb](src/resnet34.ipynb)
* [mobilenet.ipynb](src/mobilenet.ipynb)
* [alexnet.ipynb](src/alexnet.ipynb)
