import os
import pickle
import sys

import pandas as pandas
from tqdm import tqdm


def find_collisions(filename):
    print('1. Find duplicates...')
    no_duplicates = {}
    with open(filename, 'r', encoding='utf-8') as f:
        lines = f.readlines()

        for line in tqdm(lines):
            line = line[:-1]
            if line.find('.') != -1:
                line = line[::-1]
                label = line[:line.find('.')][::-1].lower()
                line = line[::-1]
                sha = line[:line.find('\t')]
                # check the presence of duplicates in the list (same sha1 address in multiple lines)
                if sha not in no_duplicates.keys():
                    no_duplicates[sha] = [label]
                else:  # collect all the different extensions which occur for the same sha1
                    if label not in no_duplicates[sha]:
                        no_duplicates[sha].append(label)
    return no_duplicates


def save_filenames(no_duplicates):
    print('\n2. Saving single-labeled files...')
    multiple = 0
    total = 0
    for sha in no_duplicates.keys():
        if len(no_duplicates[sha]) > 1:
            multiple += 1
        total += 1

    import csv
    print('Multi-labeled files are ' + str(
        round((multiple / total) * 100, 2)) + '% of the dataset.')

    with open('files_list/single_labels.csv', 'w', newline='',
              encoding='utf-8') as f:  # save only single labeled files
        writer = csv.writer(f)
        for sha in no_duplicates.keys():
            if len(no_duplicates[sha]) == 1:
                writer.writerow([sha] + no_duplicates[sha])

    return


def find_frequencies():
    print('\n3. Save label frequencies...')
    data = pandas.read_csv('files_list/single_labels.csv', names=['sha', 'label'], encoding='utf-8')

    total = data.shape[0]
    labels = data.label.unique()
    print('Number of files: ' + str(total))
    print('Number of extensions: ' + str(len(labels)))

    labels_freq_df = pandas.DataFrame(columns=['label', 'n_examples', 'frequency'])

    # Calculate for each extension the number of samples and the relative frequency
    for lab in tqdm(labels):
        n_examples = data.loc[data['label'] == lab].shape[0]
        frequency = n_examples / total
        labels_freq_df = labels_freq_df.append(
            {'label': lab, 'n_examples': int(n_examples), 'frequency': frequency},
            ignore_index=True)

    labels_freq_df = labels_freq_df.sort_values('n_examples', ascending=False)

    if not os.path.exists('results'):
        os.mkdir('results')

    labels_freq_df.to_csv('results/class_frequencies.csv', index=False)

    # save extension frequencies to a pickle file
    with open('results/class_frequencies.pkl', 'wb') as f:  # save the frequencies values
        pickle.dump(labels_freq_df, f)


def find_non_rare_extensions():
    print('\n4. Save non rare extensions...')
    # read extension frequencies
    with open('results/class_frequencies.pkl', 'rb') as f:  # save the frequency values
        extensions_frequencies = pickle.load(f)

    # Select extensions with a frequency greater than 10^-5
    selected_extensions = []
    for i in range(extensions_frequencies.shape[0]):
        row = extensions_frequencies.iloc[i]
        if round(row['frequency'], 5) > 0:
            selected_extensions.append(row['label'])

    with open('results/non_rare_extensions.pkl', 'wb') as f:  # save the frequency values
        pickle.dump(selected_extensions, f)


if __name__ == '__main__':
    filenames = find_collisions(sys.argv[1])
    save_filenames(filenames)
    find_frequencies()
    find_non_rare_extensions()
