import gzip
import os
import random
import shutil
import string
import subprocess
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
from PIL import Image, ImageFont, ImageDraw

from config import IMAGE_SIZE, GITHUB_DATASET_PATH

SYMBOLS_ALPHABET = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'

EXE = shutil.which("github-linguist")
GIT = shutil.which("git")


def linguist(path: Path, rtype: bool = False) -> Union[str, List[Tuple[str, str]]]:
    """runs Github Linguist Ruby script"""

    if not EXE:
        raise ImportError("GitHub Linguist not found, did you install it per README?")

    path = Path(path).expanduser()

    ret = subprocess.check_output([EXE, str(path)], universal_newlines=True).split("\n")

    if len(ret[3].split()) > 1:
        return " ".join(ret[3].split()[1:])
    else:
        return 'No language found!'


def render_text(text: str, filename: Path, font: ImageFont):
    img = Image.new('RGB', (IMAGE_SIZE, IMAGE_SIZE), (0, 0, 0))

    d = ImageDraw.Draw(img)
    d.text((2, 2), text, font=font, fill=(255, 255, 255), spacing=1)

    img.save(filename, 'PNG')


def get_source_code(sha):
    content = ''
    try:
        f = open(os.path.join(GITHUB_DATASET_PATH, sha[:2], sha[2:4], sha), 'rb')
    except FileNotFoundError:
        print('No File')
        return content
    except IsADirectoryError:
        print('Invalid Path')
        return content
    except UnicodeEncodeError:
        print('UnicodeEncodeError')
        return content

    try:
        content = gzip.decompress(f.read()).decode(encoding='ascii', errors='ignore')
        f.close()
        return content
    except OSError:
        print('OS_Error')
        content = f.read().decode(encoding='ascii', errors='ignore')
        f.close()
    except EOFError:
        print('EOFError')
    return content


def get_source_code_from_file(filename):
    content = ''
    try:
        f = open(filename, 'rb')
    except FileNotFoundError:
        print('No File')
        return content
    except IsADirectoryError:
        print('Invalid Path')
        return content
    except UnicodeEncodeError:
        print('UnicodeEncodeError')
        return content

    try:
        content = gzip.decompress(f.read()).decode(encoding='ascii', errors='ignore')
        f.close()
        return content
    except OSError:
        print('OS_Error')
        content = f.read().decode(encoding='ascii', errors='ignore')
        f.close()
    except EOFError:
        print('EOFError')
    return content


def scramble_all_x(letter):
    if letter.isalpha():
        return 'x'
    elif letter.isdigit():
        return 'x'
    elif letter in SYMBOLS_ALPHABET:
        return 'x'
    else:
        return letter


def scramble_alpha(letter):
    if letter.isalpha():
        if letter.islower():
            return random.choice(string.ascii_lowercase)
        else:
            return random.choice(string.ascii_uppercase)
    else:
        return letter


def scramble_num(letter):
    if letter.isdigit():
        return str(random.randint(0, 9))
    else:
        return letter


def scramble_sym(letter):
    if letter in SYMBOLS_ALPHABET:
        return random.choice(SYMBOLS_ALPHABET)
        # return random.choice(SYMBOLS_ALPHABET.replace(letter, ''))
    else:
        return letter


def scramble_alphanum(letter):
    if letter.isalpha():
        if letter.islower():
            return random.choice(string.ascii_lowercase)
        else:
            return random.choice(string.ascii_uppercase)
    elif letter.isdigit():
        return str(random.randint(0, 9))
    else:
        return letter


def scramble_alphasym(letter):
    if letter.isalpha():
        if letter.islower():
            return random.choice(string.ascii_lowercase)
        else:
            return random.choice(string.ascii_uppercase)
    elif letter in SYMBOLS_ALPHABET:
        return random.choice(SYMBOLS_ALPHABET)
    else:
        return letter


def scramble_numsym(letter):
    if letter.isdigit():
        return str(random.randint(0, 9))
    elif letter in SYMBOLS_ALPHABET:
        return random.choice(SYMBOLS_ALPHABET)
    else:
        return letter


def scramble_alphanumsym(letter):
    if letter.isalpha():
        if letter.islower():
            return random.choice(string.ascii_lowercase)
        else:
            return random.choice(string.ascii_uppercase)
    elif letter.isdigit():
        return str(random.randint(0, 9))
    elif letter in SYMBOLS_ALPHABET:
        return random.choice(SYMBOLS_ALPHABET)
    else:
        return letter


def avoid_null(value):
    if np.isnan(value):
        return 0
    return value


def compute_metrics_per_class(conf_matr, learner):
    column_sums = conf_matr.sum(axis=0)
    rows_sums = conf_matr.sum(axis=1)

    for i in range(len(learner.data.classes)):
        precision_i = round(avoid_null(conf_matr[i][i] / column_sums[i]), 3)
        recall_i = round(avoid_null(conf_matr[i][i] / rows_sums[i]), 3)
        f1_i = round(avoid_null(2 * precision_i * recall_i / (precision_i + recall_i)), 3)

        print(learner.data.classes[i] + ', ' + str(precision_i) + ', ' + str(recall_i) + ', ' + str(f1_i))


def compute_micro_macro_metrics(conf_matr, learner):
    n_classes = len(learner.data.classes)
    column_sums = conf_matr.sum(axis=0)
    rows_sums = conf_matr.sum(axis=1)

    precision_sum = 0
    recall_sum = 0
    f1_sum = 0
    for i in range(n_classes):
        precision_i = round(avoid_null(conf_matr[i][i] / (column_sums[i])), 3)
        recall_i = round(avoid_null(conf_matr[i][i] / (rows_sums[i])), 3)
        f1_i = round(avoid_null(2 * precision_i * recall_i / (precision_i + recall_i)), 3)

        precision_sum += precision_i
        recall_sum += recall_i
        f1_sum += f1_i

    print("Macro Avg Precision: " + str(round(precision_sum / n_classes, 3)))
    print("Macro Avg Recall: " + str(round(recall_sum / n_classes, 3)))
    print("Macro Avg F1: " + str(round(f1_sum / n_classes, 3)))

    tp_sum = 0
    for i in range(len(learner.data.classes)):
        tp_sum += conf_matr[i][i]

    micro_avg_precision = round(tp_sum / column_sums.sum(), 3)
    micro_avg_recall = round(tp_sum / rows_sums.sum(), 3)
    micro_avg_f1 = round(2 * micro_avg_precision * micro_avg_recall / (micro_avg_precision + micro_avg_recall), 3)
    print("\nMicro Avg Precision: " + str(micro_avg_precision))
    print("Micro Avg Recall: " + str(micro_avg_recall))
    print("Micro Avg F1: " + str(micro_avg_f1))
